// Licensed to the Apache Software Foundation (ASF) under one
// or more contributor license agreements.  See the NOTICE file
// distributed with this work for additional information
// regarding copyright ownership.  The ASF licenses this file
// to you under the Apache License, Version 2.0 (the
// "License"); you may not use this file except in compliance
// with the License.  You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

const x = window.scrollX;
const y = window.scrollY;

const width = Math.max(
    window.innerWidth || 0,
    document.documentElement.clientWidth || 0,
    document.getElementsByTagName("body").clientWidth || 0
);

const height = Math.max(
    window.innerHeight || 0,
    document.documentElement.clientHeight || 0,
    document.getElementsByTagName("body").clientHeight || 0
);

return {
    'x': x,
    'y': y,
    'w': width,
    'h': height
};
